@@ -35,17 +35,6 @@ static xmlXPathCompExprPtr models_xpath;
 static xmlXPathCompExprPtr layouts_xpath;
 static xmlXPathCompExprPtr option_groups_xpath;
 
-
-#define XKBCR_MODEL_PATH "/xkbConfigRegistry/modelList/model"
-#define XKBCR_LAYOUT_PATH "/xkbConfigRegistry/layoutList/layout"
-#define XKBCR_VARIANT_PATH XKBCR_LAYOUT_PATH "/variantList/variant"
-#define XKBCR_GROUP_PATH "/xkbConfigRegistry/optionList/group"
-#define XKBCR_OPTION_PATH XKBCR_GROUP_PATH "/option"
-
-#define XML_TAG_DESCR "description"
-#define XML_TAG_SHORT_DESCR "shortDescription"
-#define XML_TAG_VENDOR "vendor"
-
 // gettext domain for translations
 #define XKB_DOMAIN "xkeyboard-config"
 
@@ -54,9 +43,6 @@ enum {
 	PROP_ENGINE,
 };
 
-#define xkl_config_registry_is_initialized(config) \
-  ( xkl_config_registry_priv(config,xpath_context) != NULL )
-
 static gboolean
 xkl_xml_find_config_item_child(xmlNodePtr iptr, xmlNodePtr * ptr)
 {
@@ -89,6 +75,7 @@ xkl_find_nonlocalized_element(xmlNodePtr ptr, const gchar * tag_name)
 {
 	xmlNodePtr found_element = NULL;
 
+	/* Look through all siblings, trying to find a node with proper name */
 	while (ptr != NULL) {
 		char *node_name = (char *) ptr->name;
 		if (ptr->type != XML_TEXT_NODE) {
@@ -96,8 +83,10 @@ xkl_find_nonlocalized_element(xmlNodePtr ptr, const gchar * tag_name)
 
 			if (lang == NULL) {	/* No language specified */
 				if (!g_ascii_strcasecmp
-				    (node_name, tag_name))
+				    (node_name, tag_name)) {
 					found_element = ptr;
+					break;
+				}
 			}
 		}
 		ptr = ptr->next;
@@ -106,12 +95,60 @@ xkl_find_nonlocalized_element(xmlNodePtr ptr, const gchar * tag_name)
 }
 
 static gboolean
+xkl_item_populate_optional_array(XklConfigItem * item, xmlNodePtr ptr,
+				 const gchar list_tag[],
+				 const gchar element_tag[],
+				 const gchar property_name[])
+{
+	xmlNodePtr top_list_element =
+	    xkl_find_nonlocalized_element(ptr, list_tag), element_ptr;
+	gint n_elements, idx;
+	gchar **elements = NULL;
+
+	if (top_list_element == NULL || top_list_element->children == NULL)
+		return FALSE;
+
+	n_elements = 0;
+
+	/* First, count countries */
+	element_ptr = top_list_element->children;
+	while (NULL !=
+	       (element_ptr =
+		xkl_find_nonlocalized_element(element_ptr, element_tag))) {
+		n_elements++;
+		element_ptr = element_ptr->next;
+	}
+
+	if (n_elements == 0)
+		return FALSE;
+
+	elements = g_new0(gchar *, n_elements + 1);
+	/* Then, actually, populate the list */
+	element_ptr = top_list_element->children;
+	for (idx = 0;
+	     NULL != (element_ptr =
+		      xkl_find_nonlocalized_element
+		      (element_ptr, element_tag));
+	     element_ptr = element_ptr->next, idx++) {
+		elements[idx] =
+		    g_strdup((const char *) element_ptr->
+			     children->content);
+	}
+
+	g_object_set_data_full(G_OBJECT(item),
+			       property_name, elements, (GDestroyNotify)
+			       g_strfreev);
+	return TRUE;
+}
+
+gboolean
 xkl_read_config_item(XklConfigRegistry * config, xmlNodePtr iptr,
 		     XklConfigItem * item)
 {
 	xmlNodePtr name_element, ptr;
 	xmlNodePtr desc_element = NULL, short_desc_element =
 	    NULL, vendor_element = NULL;
+
 	gchar *vendor = NULL;
 
 	*item->name = 0;
@@ -119,6 +156,8 @@ xkl_read_config_item(XklConfigRegistry * config, xmlNodePtr iptr,
 	*item->description = 0;
 
 	g_object_set_data(G_OBJECT(item), XCI_PROP_VENDOR, NULL);
+	g_object_set_data(G_OBJECT(item), XCI_PROP_COUNTRY_LIST, NULL);
+	g_object_set_data(G_OBJECT(item), XCI_PROP_LANGUAGE_LIST, NULL);
 
 	if (!xkl_xml_find_config_item_child(iptr, &ptr))
 		return FALSE;
@@ -136,18 +175,6 @@ xkl_read_config_item(XklConfigRegistry * config, xmlNodePtr iptr,
 	vendor_element =
 	    xkl_find_nonlocalized_element(ptr, XML_TAG_VENDOR);
 
-	if (vendor_element != NULL && vendor_element->children != NULL) {
-		vendor =
-		    g_strdup((const char *) vendor_element->children->
-			     content);
-		g_object_set_data_full(G_OBJECT(item), XCI_PROP_VENDOR,
-				       vendor, g_free);
-	}
-
-	/*
-	 * Actually, here we should have some code to find 
-	 * the correct localized description...
-	 */
 	if (name_element != NULL && name_element->children != NULL)
 		strncat(item->name,
 			(char *) name_element->children->content,
@@ -168,6 +195,22 @@ xkl_read_config_item(XklConfigRegistry * config, xmlNodePtr iptr,
 				 (const char *) desc_element->children->
 				 content), XKL_MAX_CI_DESC_LENGTH - 1);
 	}
+
+	if (vendor_element != NULL && vendor_element->children != NULL) {
+		vendor =
+		    g_strdup((const char *) vendor_element->children->
+			     content);
+		g_object_set_data_full(G_OBJECT(item), XCI_PROP_VENDOR,
+				       vendor, g_free);
+	}
+
+	xkl_item_populate_optional_array(item, ptr, XML_TAG_COUNTRY_LIST,
+					 XML_TAG_ISO3166ID,
+					 XCI_PROP_COUNTRY_LIST);
+	xkl_item_populate_optional_array(item, ptr, XML_TAG_LANGUAGE_LIST,
+					 XML_TAG_ISO639ID,
+					 XCI_PROP_LANGUAGE_LIST);
+
 	return TRUE;
 }
 
@@ -191,7 +234,7 @@ xkl_config_registry_foreach_in_nodeset(XklConfigRegistry * config,
 	}
 }
 
-static void
+void
 xkl_config_registry_foreach_in_xpath(XklConfigRegistry * config,
 				     xmlXPathCompExprPtr
 				     xpath_comp_expr,
@@ -214,7 +257,7 @@ xkl_config_registry_foreach_in_xpath(XklConfigRegistry * config,
 	}
 }
 
-static void
+void
 xkl_config_registry_foreach_in_xpath_with_param(XklConfigRegistry
 						* config,
 						const gchar *
